# This module load the outputs saved for each individual forecast system
# and creates the multimodel ensemble

## TO DO: 
## Remove empty folders after cleaning files
## Return probabilities

source("modules/Loading/R/dates2load.R")
source("modules/Loading/R/get_timeidx.R")
source("modules/Loading/R/check_latlon.R")
source('modules/Multimodel/load_multimodel.R')
source('modules/Multimodel/load_multimodel_splitted.R')
source('modules/Multimodel/dims_multimodel.R')
source('modules/Multimodel/build_multimodel.R')
source('modules/Multimodel/clean_multimodel.R')
library(startR)
library(CSTools)
library(yaml)
recipe <- read_yaml("recipe_tas.yml")
source("modules/Saving/R/get_dir.R")
source("modules/Loading/Loading.R")
recipe$Run$output_dir <- paste0(recipe$Run$output_dir, "recipe_tas_20240604104736/")
#recipe$Run$output_dir <- paste0(recipe$Run$output_dir, "recipe_tas_20240531165840/")
# DON'T KNOW HOW TO SOLVE LOGGER:
recipe$Run$Loglevel <- 'INFO'
recipe$Run$logfile <- "/esarchive/scratch/nperez/git4//recipe_tas_20240531165840/logs/main.log" 
recipe$Run$logger <-  logger <- log4r::logger(threshold = recipe$Run$Loglevel,  
                            appenders = list(console_appender(layout = default_log_layout()),
                                             file_appender(logfile, append = TRUE,   
                                                           layout = default_log_layout())))

Multimodel <- function(recipe) {
  
  # recipe: auto-s2s recipe as provided by read_yaml
  ncores <- recipe$Analysis$ncores
  cross.method <- recipe$Analysis$cross.method
  if (is.null(cross.method)) {
    cross.method <- 'leave-one-out'
  }
  categories <- recipe$Analysis$Workflow$Probabilities$percentiles
  categories <- lapply(categories, function (x) { 
                       sapply(x, function(y) {
                              round(eval(parse(text = y)),2)})})
  na.rm <- recipe$Analysis$remove_NAs 
  skill_metrics <- list()
  requested_metrics <- strsplit(recipe$Analysis$Workflow$Skill$metric,
                                ", | |,")[[1]]
  # Load data splitting by system individual members:
  if (recipe$Analysis$Datasets$Multimodel$approach == 'pooled') { 
    data_aux <- load_multimodel(recipe)
    datos <- append(list(obs = data_aux$obs$data), lapply(data_aux$hcst, function(x) x$data))
    names(datos)[-1] <- sapply(recipe$Analysis$Datasets$System, '[[', 'name')
    # Compute crps /crpss and return forecast for visualization purposes   
    if (any(c('crps', 'crpss') %in% requested_metrics)) {
      crps <- Apply(datos, target_dims = c('syear', 'ensemble'), 
                    fun = function(obs, ...) {
                      res <- abind(..., along = 2)
                      names(dim(res)) <- names(dim(obs))
                      obs <- Subset(obs, along = 'ensemble', indices = 1, drop = 'selected')
                      s2dv:::.CRPS(exp = res, obs = obs, dat_dim = NULL, 
                         time_dim = 'syear', memb_dim = 'ensemble')},
                         ncores = ncores)$output1
      skill_metrics$crps <- crps  
      # Build the reference forecast:
      sdate_dim <- dim(datos[[1]])['syear']
      cross <- CSTools:::.make.eval.train.dexes(eval.method = cross.method,
                                                amt.points = sdate_dim,
                                                amt.points_cor = NULL) 
      tmp <- array(1:length(cross), c(syear = length(cross)))
      ref_clim <- Apply(list(datos$obs, tmp),
                        target_dims = list(c('ensemble', 'syear'), NULL),
                        function(x, y) {
                          Subset(x, along = 'syear', indices = cross[[y]]$train.dexes, drop = T)},
                        ncores = ncores, output_dims = 'ensemble')$output1
      datos <- append(list(ref = ref_clim), datos)
      crps_clim <- Apply(datos[1:2], target_dims = c('syear', 'ensemble'),
                         fun = function(ref, obs) {
                           obs <- Subset(obs, along = 'ensemble', indices = 1, drop = 'selected')
                           s2dv:::.CRPS(exp = ref, obs = obs, dat_dim = NULL,
                           time_dim = 'syear', memb_dim = 'ensemble')},
                         ncores = ncores)$output1
      skill_metrics$crps_clim <- crps_clim


      crpss <- Apply(datos, target_dims = c('syear', 'ensemble'), 
                     fun = function(ref, obs, ...) {
                       res <-  abind(..., along = 2)
                       names(dim(res)) <- names(dim(obs))
                       obs <- Subset(obs, along = 'ensemble', indices = 1, drop = 'selected')
                       CRPSS(exp = res, obs = obs, ref = ref,
                            memb_dim = 'ensemble', Fair = FALSE,
                            time_dim = 'syear', clim.cross.val = FALSE)},
                     ncores = ncores)
      skill_metrics$crpss <- crpss$crpss
      skill_metrics$crpss_significance <- crpss$sign
    }
    # What about spread or spread-to error ratio?
    rm(datos)    
  } 
  # Always load ensemble mean and probabilities:
  source("modules/Multimodel/load_multimodel_mean.R")
  datos <- load_multimodel_mean(recipe)  
  # Building the multi-model
  multimodel_aux <- build_multimodel(data, recipe)
  data <- multimodel_aux$data
  prob <- multimodel_aux$prob
  rm(multimodel_aux)

  # Saving multimodel
  if (recipe$Analysis$Workflow[[recipe$Analysis$Datasets$Multimodel$createFrom]]$save != 'none') {
    save_forecast(recipe = recipe,
                  data_cube = data$hcst,
                  outdir = NULL,
                  type = 'hcst')
    if (!is.null(data$fcst)) {
      save_forecast(recipe = recipe, 
                    data_cube = data$fcst,
                    outdir = NULL,
                    type = 'fcst')
    }
    save_observations(recipe = recipe, 
                      data_cube = data$obs,
                      outdir = NULL)
  }
  
  # Removing the temporary data of models in case user only requests multimodel
  clean_multimodel(recipe,
                   files_hcst, 
                   files_fcst, 
                   files_obs)
  
  return(list(data = data, 
              prob = prob))
  
}
